from __future__ import annotations

from django import forms
from django.db.models.fields.files import FieldFile, ImageField

from .app_settings import app_settings


class DragDropFileInput(forms.ClearableFileInput):
    template_name = "django_chunk_file_upload/forms/widgets/drag_drop_input.html"

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context["widget"]["attrs"]["required"] = False
        context["widget"]["attrs"]["hidden"] = True
        context["widget"]["attrs"]["data-id"] = "dropzone"
        if value and isinstance(value, (FieldFile, ImageField)):
            context["widget"]["attrs"]["data-value"] = value.instance.pk
        return context

    @property
    def media(self):
        return forms.Media(
            css={"all": app_settings.css},
            js=app_settings.js,
        )


class DragDropFileField(forms.FileField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("widget", DragDropFileInput())
        super().__init__(*args, **kwargs)


class MultipleDragDropFileInput(DragDropFileInput):
    allow_multiple_selected = True


class MultipleDragDropFileField(forms.FileField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("widget", MultipleDragDropFileInput())
        super().__init__(*args, **kwargs)

    def clean(self, data, initial=None):
        single_file_clean = super().clean
        if isinstance(data, (list, tuple)):
            result = [single_file_clean(d, initial) for d in data]
        else:
            result = single_file_clean(data, initial)
        return result



from __future__ import annotations

from dataclasses import asdict, dataclass, field, fields
from re import match
from typing import Any, Union
from uuid import UUID

from django.core.files.uploadedfile import (
    InMemoryUploadedFile,
    TemporaryUploadedFile,
)
from django.utils.translation import gettext_lazy as _

from .constants import TypeChoices
from .optimize import MapOptimizer
from .utils import (
    get_file_extension,
    get_file_path,
    get_filename,
    get_save_file_path,
    make_uuid,
)


@dataclass(kw_only=True)
class BaseFile:
    """Base File"""

    common_types = {
        TypeChoices.ARCHIVE: (
            ".bz",
            ".bz2",
            ".gz",
            ".jar",
            ".rar",
            ".tar",
            ".zip",
            ".7z",
        ),
        TypeChoices.AUDIO: (
            ".aac",
            ".mid",
            ".midi",
            ".mp3",
            ".oga",
            ".opus",
            ".wav",
            ".weba",
            ".3gp",
            ".3g2",
        ),
        TypeChoices.BINARY: (".bin",),
        TypeChoices.DOCUMENT: (
            ".abw",
            ".arc",
            ".odp",
            ".odt",
            ".pdf",
            ".md",
        ),
        TypeChoices.FONT: (
            ".eot",
            ".otf",
            ".ttf",
            ".woff",
            ".woff2",
        ),
        TypeChoices.HYPERTEXT: (
            ".html",
            ".HTM",
        ),
        TypeChoices.IMAGE: (
            ".apng",
            ".avif",
            ".bmp",
            ".gif",
            ".ico",
            ".jpeg",
            ".jpg",
            ".png",
            ".svg",
            ".tif",
            ".tiff",
            ".webp",
        ),
        TypeChoices.JSON: (
            ".json",
            ".jsonld",
        ),
        TypeChoices.MICROSOFT_WORD: (
            ".doc",
            ".docx",
        ),
        TypeChoices.MICROSOFT_POWERPOINT: (
            ".ppt",
            ".pptx",
        ),
        TypeChoices.MICROSOFT_EXCEL: (
            ".xls",
            ".xlsx",
        ),
        TypeChoices.SEPARATED: (
            ".csv",
            ".tsv",
        ),
        TypeChoices.TEXT: ("txt",),
        TypeChoices.VIDEO: (
            ".avi",
            ".mp4",
            ".mpeg",
            ".ogg",
            ".mp2t",
            ".webm",
            ".3gpp",
            ".3gpp2",
        ),
        TypeChoices.XML: ("xml",),
    }
    _accepted_mime_types: list = field(default_factory=lambda: [".*/*"])
    _id: UUID = None
    _user: Any = None
    _file: Any = None
    _type: TypeChoices = None
    _root_dir: str = None
    _path: str = None
    _extension: str = None
    _upload_to: str = None
    _message: str = None
    checksum: str = None
    chunk_from: str = None
    chunk_size: str = None
    chunk_to: str = None
    eof: bool = False
    mimetype: str = None
    name: str = None
    size: str = None

    @property
    def id(self) -> Any:
        if self._id is None:
            self._id = make_uuid(user=self.user, checksum=self.checksum)
        return self._id

    @property
    def file(self) -> Union[InMemoryUploadedFile, TemporaryUploadedFile, None]:
        return self._file

    @property
    def filename(self) -> str:
        return get_filename(getattr(self.file, "name", "")) or ""

    @property
    def repl_filename(self) -> str:
        return str(self.id)

    @property
    def user(self) -> Any:
        return self._user

    @property
    def path(self) -> str:
        if self._path is None:
            self._path = get_file_path(
                self.repl_filename + self.extension, self._upload_to
            )
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        self._path = value

    @property
    def save_path(self) -> str:
        return get_save_file_path(self.path)

    @property
    def extension(self) -> str:
        if self._extension is None:
            self._extension = get_file_extension(self.file.name)
        return self._extension

    @extension.setter
    def extension(self, value: str) -> None:
        self._extension = value

    @property
    def type(self) -> TypeChoices:
        if self._type is None:
            self._type = self._get_type(self.extension)
        return self._type

    @property
    def message(self) -> str:
        if self._message is None:
            self._message = _("Uploading file, please wait a moment.")
            if self.eof:
                self._message = _("File upload is completed.")
        return self._message

    @message.setter
    def message(self, value: str) -> None:
        self._message = value

    @classmethod
    def model_fields_set(cls) -> set:
        return {obj.name for obj in fields(cls)}

    @classmethod
    def from_request(cls, request, upload_to: str = "%Y/%m/%d") -> "File":
        """Create instance from request

        Receive requests from the client using jQuery AJAX Method.
        We receive parameters using request headers, including::
            X-File-Name: File name
            X-File-Checksum: File MD5 checksum
            X-File-Chunk-From: Chunk from of File.
            X-File-Chunk-To: Chunk to of File.
            X-File-Chunk-Size: Chunk size per request.
            X-File-EOF: True is upload completed, otherwise.
            X-File-Size: Original file size.
            X-File-Mime-Type: MINE type of file.

        Reference: django_chunk_file_upload/static/js/upload.chunk.js

        Args:
          request: AJAX request from client
          upload_to: Server upload dir.

        Returns:
          The File Metadata instance.
        """

        def to_private_attrs():
            user = None
            if request.user and request.user.is_authenticated:
                user = request.user

            file = request.FILES.get("file")
            pk = make_uuid(user=user, checksum=request.headers.get("x-file-checksum"))
            return {
                "_id": pk,
                "_extension": get_file_extension(file.name) if file else None,
                "_file": file,
                "_user": user,
                "_upload_to": upload_to,
            }

        kwargs = {}
        model_fields = {
            model_field
            for model_field in cls.model_fields_set()
            if not model_field.startswith("_")
        }

        for k, v in request.headers.items():
            if str(k).lower().startswith("x-file-"):
                param = str(k)[7:].replace("-", "_").lower()
                if param in model_fields:
                    if param == "eof":
                        v = bool(str(v).lower() in ["true", "1", "yes", "on"])

                    kwargs[param] = v

        kwargs.update(to_private_attrs())
        ret = cls(**kwargs)
        return ret

    def _get_type(self, extension: str) -> TypeChoices:
        for common_type, extensions in self.common_types.items():
            if str(extension).lower() in extensions:
                return common_type
        return TypeChoices.__empty__

    def is_valid(self) -> bool:
        if not (self.checksum and self.mimetype and self.file):
            return False

        for pattern in self._accepted_mime_types:
            if match(pattern, self.mimetype):
                return True

        return False

    def to_dict(self) -> dict:
        return asdict(self)

    def to_metadata(self) -> dict:
        return self.to_response().copy()

    def to_response(self) -> dict:
        metadata = {k: v for k, v in self.to_dict().items() if not k.startswith("_")}
        metadata["message"] = str(self.message)
        metadata["name"] = self.filename
        return metadata

    def write(self, mode: str = "ab+"):
        with open(self.save_path, mode) as fp:
            for chunk in self.file.chunks(self.chunk_size):
                fp.write(chunk)

    def optimize(self, instance):
        optimizer = MapOptimizer.get(self.type, None)
        if optimizer and isinstance(optimizer, type):
            optimizer = optimizer(instance, self)
            optimizer.optimize()


@dataclass(kw_only=True)
class File(BaseFile):
    """File"""

    _accepted_mime_types: list = field(default_factory=lambda: [".*/*"])


@dataclass(kw_only=True)
class AudioFile(File):
    """Audio File Metadata"""

    _accepted_mime_types: list = field(default_factory=lambda: [r"audio/*"])


@dataclass(kw_only=True)
class ArchiveFile(File):
    """Archive File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [
            r"application/x-bzip",
            r"application/x\-bzip2",
            r"application/gzip",
            r"application/x\-gzip",
            r"application/java\-archive",
            r"application/vnd\.rar",
            r"application/x\-tar",
            r"application/zip",
            r"x\-zip\-compressed",
            r"application/x\-7z\-compressed",
        ]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.ARCHIVE


@dataclass(kw_only=True)
class BinaryFile(File):
    """Binary File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [r"application/octet\-stream"]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.BINARY


@dataclass(kw_only=True)
class DocumentFile(File):
    """Document File Metadata"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [
            r"application/x\-abiword",
            r"application/x\-freearc",
            r"application/msword",
            r"application/vnd\.openxmlformats\-officedocument\.wordprocessingml\.document",
            r"application/vnd\.oasis\.opendocument\.presentation",
            r"application/vnd\.oasis\.opendocument\.text",
            r"application/pdf",
            r"text/markdown",
        ]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.DOCUMENT


@dataclass(kw_only=True)
class FontFile(File):
    """Font File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [r"application/vnd\.ms\-fontobject", r"font/*"]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.FONT


@dataclass(kw_only=True)
class HyperTextFile(File):
    """HyperText Markup Language File Metadata"""

    _accepted_mime_types: list = field(default_factory=lambda: [r"text/html"])

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.HYPERTEXT


@dataclass(kw_only=True)
class ImageFile(File):
    """Image File"""

    _accepted_mime_types: list = field(default_factory=lambda: [r"image/*"])

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.IMAGE


@dataclass(kw_only=True)
class JSONFile(File):
    """JSON File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [r"application/json", r"application/ld\+json"]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.JSON


@dataclass(kw_only=True)
class MicrosoftWordFile(File):
    """Microsoft Word File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [
            r"application/msword",
            r"application/vnd\.openxmlformats\-officedocument\.wordprocessingml\.document",
        ]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.MICROSOFT_WORD


@dataclass(kw_only=True)
class MicrosoftPowerPointFile(File):
    """Microsoft PowerPoint File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [
            r"application/vnd\.ms\-powerpoint",
            r"application/vnd\.openxmlformats\-officedocument\.presentationml\.presentation",
        ]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.MICROSOFT_POWERPOINT


@dataclass(kw_only=True)
class MicrosoftExcelFile(File):
    """Microsoft Excel File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [
            r"application/vnd\.ms\-excel",
            r"application/vnd\.openxmlformats\-officedocument\.spreadsheetml\.sheet",
        ]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.MICROSOFT_EXCEL


@dataclass(kw_only=True)
class SeparatedFile(File):
    """Separated File"""

    _accepted_mime_types: list = field(default_factory=lambda: [r"text/csv"])

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.SEPARATED


@dataclass(kw_only=True)
class XMLFile(File):
    """XML File"""

    _accepted_mime_types: list = field(
        default_factory=lambda: [
            r"application/xml",
            r"text/xml",
            r"application/atom\+xml",
        ]
    )

    @property
    def type(self) -> TypeChoices:
        return TypeChoices.XML



from __future__ import annotations

import hashlib
import json
import os
from io import BytesIO
from typing import Union
from uuid import UUID

from django.conf import settings
from django.core.files.uploadedfile import (
    InMemoryUploadedFile,
    TemporaryUploadedFile,
)
from django.utils import timezone


def _attr_to_str(*args, **kwargs) -> str:
    return ":".join([str(arg) for arg in args] + [json.dumps(kwargs, default=str)])


def make_md5_hash(*args, **kwargs) -> str:
    return hashlib.md5(_attr_to_str(*args, **kwargs).encode("utf-8")).hexdigest()


def make_uuid(*args, **kwargs) -> UUID:
    return UUID(hex=make_md5_hash(*args, **kwargs))


def create_dir(dir_path: str) -> None:
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def get_dir(upload_to: str = None) -> str:
    media_root = str(settings.MEDIA_ROOT) if settings.MEDIA_ROOT else ""
    root_dir = os.path.join(
        media_root,
        (
            timezone.now().strftime(upload_to)
            if isinstance(upload_to, str)
            else timezone.now().strftime("%Y/%m/%d")
        ),
    )
    return root_dir


def safe_remove_file(fp: str) -> None:
    try:
        os.remove(fp)
    except OSError:
        pass


def get_filename(fp: str, remove_extension: bool = False) -> str:
    _, filename = os.path.split(fp)
    if remove_extension:
        filename = os.path.splitext(filename)[0]
    return filename


def get_file_extension(fp: str) -> str:
    _, extension = os.path.splitext(fp)
    return extension


def get_save_file_path(fp: str, upload_to: str = None) -> str:
    filename = get_filename(fp)
    root_dir = get_dir(upload_to)
    create_dir(root_dir)
    fp = os.path.join(root_dir, filename)
    return fp


def get_paths(fp: str, upload_to: str = "") -> tuple[str, str]:
    save_fp = get_save_file_path(fp, upload_to)
    media_root = str(settings.MEDIA_ROOT) if settings.MEDIA_ROOT else ""
    if media_root:
        fp = save_fp.split(media_root)[-1]
    else:
        fp = save_fp

    if fp[0] == "/":
        fp = fp[1:]

    return save_fp, fp


def get_file_path(filename: str, upload_dir: str = "") -> str:
    _save_fp, fp = get_paths(filename, upload_dir)
    return fp


def join_file_path(*args: str) -> str:
    return os.path.join(*args)


def handle_upload_file(file, upload_dir: str = None):
    save_fp, fp = get_paths(file.name, upload_dir)
    with open(save_fp, "wb+") as f:
        for chunk in file.chunks():
            f.write(chunk)
        return fp


def get_md5_checksum(
    fp: Union[str, bytes, InMemoryUploadedFile, TemporaryUploadedFile],
    chunk_size: int = 65536,
) -> str:
    md5hash = hashlib.md5()
    if isinstance(fp, str):
        with open(fp, "rb") as fp:
            while chunk := fp.read(chunk_size):
                md5hash.update(chunk)
    elif isinstance(fp, bytes):
        while chunk := BytesIO(fp).read(chunk_size):
            md5hash.update(chunk)
    else:
        for chunk in fp.chunks(chunk_size):
            md5hash.update(chunk)
    return md5hash.hexdigest()



from PIL import Image, UnidentifiedImageError
from PIL.JpegImagePlugin import JpegImageFile
from PIL.PngImagePlugin import PngImageFile
from PIL.WebPImagePlugin import WebPImageFile

from .app_settings import app_settings
from .constants import TypeChoices
from .utils import get_file_path


class BaseOptimizer:
    """Base Optimizer"""

    def __init__(self, instance, file, *args, **kwargs):
        self._instance = instance
        self._file = file

    @property
    def instance(self):
        return self._instance

    @property
    def file(self):
        return self._file

    def optimize(self):
        pass


class ImageOptimizer(BaseOptimizer):
    """Image Optimizer"""

    _supported_file_types = (".jpg", ".jpeg", ".png", ".webp")

    def __init__(
        self,
        instance,
        file,
        *args,
        **kwargs,
    ):
        super().__init__(instance, file, *args, **kwargs)

    def open(self) -> Image.Image | None:
        try:
            image = Image.open(self.file.file)
            return image
        except UnidentifiedImageError:
            pass

    def close(self, image: Image.Image) -> None:
        if isinstance(image, Image.Image):
            image.close()

    def optimize(self) -> None:
        resized_img = self.open()
        if not resized_img:
            return

        fm, ext = None, None
        if isinstance(resized_img, PngImageFile):
            fm, ext = "PNG", ".png"
            resized_img = resized_img.convert("P", palette=Image.ADAPTIVE)
        elif isinstance(resized_img, JpegImageFile):
            fm, ext = "JPEG", ".jpg"
        elif isinstance(resized_img, WebPImageFile):
            fm, ext = "WEBP", ".webp"

        if app_settings.image_optimizer.to_webp:
            fm, ext = "WEBP", ".webp"

        if str(ext) in self._supported_file_types:
            resized_img = self.resize(resized_img)
            orig_save_path = self.file.save_path
            filename = self.file.repl_filename + ext
            self.file.path = get_file_path(filename, self.file._upload_to)
            resized_img.save(
                self.file.save_path,
                fm,
                optimize=True,
                quality=app_settings.image_optimizer.quality,
                compress_level=app_settings.image_optimizer.compress_level,
            )
            if orig_save_path != self.file.save_path:
                self.instance.file.delete()
                self.file.extension = ext

        self.close(resized_img)

    def resize(self, image: Image.Image) -> Image.Image:
        """Resize image to fit with max width and max height"""

        w, h = image.size
        aspect_ratio = w / h

        if (
            w > app_settings.image_optimizer.max_width
            or h > app_settings.image_optimizer.max_height
        ):
            if aspect_ratio > 1:
                nw = app_settings.image_optimizer.max_width
                nh = int(app_settings.image_optimizer.max_width / aspect_ratio)
            else:
                nh = app_settings.image_optimizer.max_height
                nw = int(app_settings.image_optimizer.max_height * aspect_ratio)

            return image.resize((nw, nh), Image.LANCZOS)
        return image


MapOptimizer = {TypeChoices.IMAGE: ImageOptimizer}



# Generated by Django 5.0.7 on 2024-08-19 20:46

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_chunk_file_upload", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="filemanager",
            options={
                "ordering": ("-created_at",),
                "verbose_name": "File Manager",
                "verbose_name_plural": "File Manager",
            },
        ),
        migrations.AddField(
            model_name="filemanager",
            name="type",
            field=models.CharField(
                choices=[
                    (None, "Unknown"),
                    ("ARCHIVE", "ARCHIVE"),
                    ("AUDIO", "AUDIO"),
                    ("BINARY", "BINARY"),
                    ("DOCUMENT", "DOCUMENT"),
                    ("FONT", "FONT"),
                    ("HYPERTEXT", "HYPERTEXT"),
                    ("IMAGE", "IMAGE"),
                    ("JSON", "JSON"),
                    ("MICROSOFT_EXCEL", "MICROSOFT_EXCEL"),
                    ("MICROSOFT_POWERPOINT", "MICROSOFT_POWERPOINT"),
                    ("MICROSOFT_WORD", "MICROSOFT_WORD"),
                    ("SOURCE_CODE", "SOURCE_CODE"),
                    ("SEPARATED", "SEPARATED"),
                    ("TEXT", "TEXT"),
                    ("VIDEO", "VIDEO"),
                    ("XML", "XML"),
                ],
                default="Unknown",
                max_length=255,
                verbose_name="Type",
            ),
        ),
        migrations.AddField(
            model_name="filemanager",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="files",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="filemanager",
            name="checksum",
            field=models.CharField(max_length=255),
        ),
        migrations.RemoveField(
            model_name="filemanager",
            name="name",
        ),
        migrations.AlterUniqueTogether(
            name="filemanager",
            unique_together={("user", "checksum")},
        ),
    ]



# Generated by Django 5.0.7 on 2024-08-28 02:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("django_chunk_file_upload", "0002_alter_filemanager_options_and_more"),
    ]

    operations = [
        migrations.RenameIndex(
            model_name="filemanager",
            new_name="filemanager_checksum_idx",
            old_name="file_manager_checksum_idx",
        ),
    ]
